//=============================================================================
// Coolbits - by Sebastian Kaufel.
//
// If you like it, gift me Lego.
//=============================================================================
class Coolbits extends Object
	abstract
	native
	noexport;

var name CoolBitsVersion;

// ----------------------------------------------------------------------
// CarriageReturn()
//
// Just hooks ExtensionObject.StringToName() and is made static so you
// won't need an ExtensionObject around for String to Name convertions.
//
// Example: Class'Coolbits'.Static.StringToName( 'CoolBits' )
// ----------------------------------------------------------------------

native(1024) static final function Name StringToName(coerce string str);

// ----------------------------------------------------------------------
// CarriageReturn()
//
// Just hooks Window.CarriageReturn() and is made static so you won't
// need an Window around for getting a carriage return.
//
// Example: Class'Coolbits'.Static.CarriageReturn()
// ----------------------------------------------------------------------

native(1482) static final function string CarriageReturn();

// ----------------------------------------------------------------------
// CR() - Shortened form of CarriageReturn().
//
// Just copied out of Extension.Window and made static.
// ----------------------------------------------------------------------

static function string CR()
{
	return CarriageReturn();
}

// ----------------------------------------------------------------------
// PreRenderWindows() / PostRenderWindows()
//
// Meant as a replacement for the versions in PlayerPawnExt.
// A parameter is added to specify a drawning scale, e.g. you can
// force the windows to be renderer in 1x scale. However, if your
// screen is smaller then ( Scale * Width, Scale * Height ) the
// next lower valid Scale is used. Using Scale = 0 enables the
// default DeusEx behaviour, and enables you to have an Auto option.
// 
// To use these you need to override your PlayerPawns PreRender() and
// PostRenderFlash() to make the use of these functions in place of the
// PlayerPawnExt ones.
//
// Example: Class'Coolbits'.Static.PreRenderWindows( rootWindow, Canvas, 1 )
//					Class'Coolbits'.Static.PostRenderWindows( rootWindow, Canvas, 1 )
// ----------------------------------------------------------------------

native(4004) final static function PreRenderWindows( RootWindow RootWindow, Canvas Canvas, int Scale );
native(4005) final static function PostRenderWindows( RootWindow RootWindow, Canvas Canvas, int Scale );

// ----------------------------------------------------------------------
// TraceTextureCoolBits() - Fixed version of Actors.TraceTexture().
//
// However you need to add an Actor as the first parameter who's
// Level will be used and who's Location will be used as a default
// parameter for for Start.
//
// Example:
// Old: foreach TraceTexture(class'Actor', target, texName, texGroup, texFlags, HitLocation, HitNormal, EndTrace)
// New: foreach class'CoolBits'.static.TraceTextureCoolBits(Self, class'Actor', target, texName, texGroup, texFlags, HitLocation, HitNormal, EndTrace)
// ----------------------------------------------------------------------

native(4006) final static iterator function TraceTextureCoolBits( Actor Other, class<Actor> BaseClass, out actor Actor, out name TexName, out name TexGroup, out int Flags, out vector HitLoc, out vector HitNorm, vector End, optional vector Start, optional vector Extent );

// ----------------------------------------------------------------------
// StriHash() - Wrapper for appStrihash.
//
// Returns the case insensitive hash for a given string.
// ----------------------------------------------------------------------

native(4007) final static function int StriHash( String Str );

// ----------------------------------------------------------------------
// CopyConversation() - Copies Actors (Conversation) config variables
// ----------------------------------------------------------------------

static function CopyConversation( Actor Dest, Actor Src )
{
	Dest.BindName						= Src.BindName;					// Used to bind conversations
	Dest.BarkBindName				= Src.BarkBindName;			// Used to bind Barks!
	Dest.FamiliarName				= Src.FamiliarName;			// For display in Conversations
	Dest.UnfamiliarName			= Src.UnfamiliarName;		// For display in Conversations
	Dest.ConStartInterval		= Src.ConStartInterval;	// Amount of time required between two convos.
}

// ----------------------------------------------------------------------
// CopyDisplay() - Copies Actors (Display) config variables
// ----------------------------------------------------------------------

static function CopyDisplay( Actor Dest, Actor Src )
{
	local int i;

	// Animation variables.
	Dest.AnimSequence	= Src.AnimSequence; // Animation sequence we're playing.
	Dest.AnimFrame		= Src.AnimFrame;    // Current animation frame, 0.0 to 1.0.
	Dest.AnimRate			= Src.AnimRate;     // Animation rate in frames per second, 0=none, negative=velocity scaled.
	Dest.LODBias			= Src.LODBias;

	// Drawing effect.
	Dest.DrawType			= Src.DrawType;

	// Style for rendering sprites, meshes.
	Dest.Style				= Src.Style;

	// Other display properties.
	Dest.Sprite				= Src.Sprite;			 // Sprite texture if DrawType=DT_Sprite.
	Dest.Texture			= Src.Texture;		 // Misc texture.
	Dest.Skin					= Src.Skin;        // Special skin or enviro map texture.
	Dest.Mesh					= Src.Mesh;        // Mesh if DrawType=DT_Mesh.
	Dest.DrawScale		= Src.DrawScale;	 // Scaling factor, 1.0=normal size.
	Dest.ScaleGlow		= Src.ScaleGlow;	 // Multiplies lighting.
	Dest.AmbientGlow	= Src.AmbientGlow; // Ambient brightness, or 255=pulsing.
	Dest.Fatness			= Src.Fatness;     // Fatness (mesh distortion).

	// Display.
	Dest.bUnlit						= Src.bUnlit;           // Lights don't affect actor.
	Dest.bNoSmooth				= Src.bNoSmooth;        // Don't smooth actor's texture.
	Dest.bParticles				= Src.bParticles;       // Mesh is a particle system.
	Dest.bRandomFrame			= Src.bRandomFrame;     // Particles use a random texture from among the default texture and the multiskins textures
	Dest.bMeshEnviroMap		= Src.bMeshEnviroMap;   // Environment-map the mesh.
	Dest.bMeshCurvy				= Src.bMeshCurvy;       // Curvy mesh.
	Dest.VisibilityRadius	= Src.VisibilityRadius; // Actor is drawn if viewer is within its visibility
	Dest.VisibilityHeight	= Src.VisibilityHeight; // cylinder.  Zero=infinite visibility.

	// Not yet implemented.
	Dest.bShadowCast	= Src.bShadowCast;     // Casts shadows.

	// Multiple skin support.
	for ( i = 0; i < ArrayCount(Src.MultiSkins); i++ )
		Dest.MultiSkins[i] = Src.MultiSkins[i];

	if ( Dest.IsA('Inventory') && Src.IsA('Inventory') )
	{
		Inventory(Dest).bAmbientGlow = Inventory(Src).bAmbientGlow; // Whether to glow or not.
	}
	else if ( Dest.IsA('DeusExCarcass') && Src.IsA('DeusExCarcass') )
	{
		DeusExCarcass(Dest).Mesh2 = DeusExCarcass(Src).Mesh2;		// mesh for secondary carcass
		DeusExCarcass(Dest).Mesh3 = DeusExCarcass(Src).Mesh3;		// mesh for floating carcass
	}
}

// ----------------------------------------------------------------------
// CopyLightColor() - Copies Actors (LightColor) config variables
// ----------------------------------------------------------------------

static function CopyLightColor( Actor Dest, Actor Src )
{
	// Lighting info.
	Dest.LightBrightness = Src.LightBrightness;
	Dest.LightHue				 = Src.LightHue;
	Dest.LightSaturation = Src.LightSaturation;
}

// ----------------------------------------------------------------------
// CopyLighting() - Copies Actors (Lighting) config variables
// ----------------------------------------------------------------------

static function CopyLighting( Actor Dest, Actor Src )
{
	// Light modulation.
	Dest.LightType				= Src.LightType;

	// Spatial light effect to use.
	Dest.LightEffect			= Src.LightEffect;

	// Light properties.
	Dest.LightRadius			= Src.LightRadius;
	Dest.LightPeriod			= Src.LightPeriod;
	Dest.LightPhase				= Src.LightPhase;
	Dest.LightCone				= Src.LightCone;
	Dest.VolumeBrightness	= Src.VolumeBrightness;
	Dest.VolumeRadius			= Src.VolumeRadius;
	Dest.VolumeFog				= Src.VolumeFog;

	// Lighting.
	Dest.bSpecialLit			= Src.bSpecialLit;	 // Only affects special-lit surfaces.
	Dest.bActorShadows		= Src.bActorShadows; // Light casts actor shadows.
	Dest.bCorona					= Src.bCorona;       // Light uses Skin as a corona.
	Dest.bLensFlare				= Src.bLensFlare;    // Whether to use zone lens flare.
}

// ----------------------------------------------------------------------
// CopySound() - Copies Actors (Sound) config variables
// ----------------------------------------------------------------------

static function CopySound( Actor Dest, Actor Src )
{
	// Ambient sound.
	Dest.SoundRadius	= Src.SoundRadius;	// Radius of ambient sound.
	Dest.SoundVolume	= Src.SoundVolume;	// Volume of amient sound.
	Dest.SoundPitch		= Src.SoundPitch;		// Sound pitch shift, 64.0=none.
	Dest.AmbientSound	= Src.AmbientSound;	// Ambient sound effect.

	// Regular sounds.
	Dest.TransientSoundVolume = Src.TransientSoundVolume;
	Dest.TransientSoundRadius = Src.TransientSoundRadius;
}

// ----------------------------------------------------------------------
// GetInventoryReplacementClass()
//
// Conveniance function for using DuckTypeClass.
// Nice if you want to Replace InitialInventory.
// ----------------------------------------------------------------------

static function Class<Inventory> GetInventoryReplacementClass( Class<Inventory> OtherClass, String Package, String Prefix )
{
	local class<Inventory> ReplacementClass;

	ReplacementClass = class<Inventory>( DuckTypeClass( OtherClass, Package, Prefix ) );

	if ( ReplacementClass == None )
		return OtherClass;
	else
		return ReplacementClass;
}

// ----------------------------------------------------------------------
// StripPackage()
//
// Strips the package tag out of a given string.
// ----------------------------------------------------------------------

static function string StripPackage(string s)
{
	local int p;
	
	p = InStr(s, ".");

	if (p == -1)
		return s;

	return Right(s, Len(s) - p - 1);
}

// ----------------------------------------------------------------------
// DuckTypeClass()
//
// Returns the replacement class by dynamically loading it based on name.
// 
// Example:
//   DuckTypeClass( Class'MaxChen', "CoolBits", "CB" )
//	 Would try to load CoolBits.CBMaxChen.
// ----------------------------------------------------------------------

static function class DuckTypeClass( Class InClass, String Package, String Prefix )
{
	local string ClassName;

	if ( InClass == none )
		return none;

	ClassName = Package $ "." $ Prefix $ StripPackage(InClass $ "");

	return class(DynamicLoadObject(ClassName, class'Class'));
}

// ----------------------------------------------------------------------
// ClassInPackageHack()
//
// Hacky but fast way to decide in Mutator if it wants to replace this
// Actor again.
//
// Example:
//   ClassInPackageHack( Other, "CoolBits" )
// ----------------------------------------------------------------------

static function bool ClassInPackageHack( Class CheckClass, String Package )
{
	
	return Left( CheckClass $ "", Len(Package) ) ~= "HX";
}

// ----------------------------------------------------------------------
// BindConversationList()
//
// Can by used to bind conversation events to an arbitray actor.
// e.g. bind the DataLinks to your GameInfo. You need to set a
// proper BindName for your actor (e.g. JCDenton for DataLinks).
//
// Example:
//   if ( MissionNumber >= 0 && MissionNumber <= 9 )
//   {
//     ConListClass = "DeusExConText.ConList_Mission0" $ MissionNumber;
//   }
//   else if ( MissionNumber >= 10 && MissionNumber <= 99 )
//   {
//     ConListClass = "DeusExConText.ConList_Mission" $ MissionNumber;
//   }
//   else
//   {
//     Log( "WTF IS THIS MISSION NUMBER " $ MissionNumber );
//     return false;
//   }
//   Class'Coolbits'.Static.BindConversationList( ConListClass, Self );
// ----------------------------------------------------------------------

native(4008) final static function bool BindConversationList( String ConListClass, Actor A );

// ----------------------------------------------------------------------
// ExportTravel()
//
// Updates the Levels TravelInfo's entry about the given PlayerPawn.
// The name for the entry is PlayerPawn->PlayerReplicationInfo->PlayerName.
// You will probably want to call this in PlayerPawn.Destroy() before
// the inventory gets deleted.
//
// ULevel::SpawnPlayActor() will create and link the objects once the
// player joins again.
// You modify GameInfo.AcceptInventory(). An approach would be to check
// if the player has a NanoKeyRing and if not give the initial inventory
// and keep the old if the already has a NanoKeyRing.
//
// Warning:
//   ION modified the ExportTravel() in a way that non Actor Objects
//   will be include in the TravelInfo entry. And this is a serious
//   problem since this is prone for crashes and strange issues.
//   So be warned when using this. When it crashes your game, try
//   getting rid of non actor export variables / objects. However
//   the FlagBase is constructed in a way which the GC *really*
//   like. However removing all references to flags and flagbases helps.
//   TObjectIterator is helpful.
// ----------------------------------------------------------------------

native(4009) final static function ExportTravel( PlayerPawn PlayerPawn );

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

defaultproperties
{
	CoolBitsVersion=20140919
}
