/*=============================================================================
	CoolBits.cpp: CoolBits main cpp file.
	If you like it, gift me Lego.

	Revision history:
		* Created by Sebastian Kaufel

	To compile this you need to make some new friends!
	(e.g. Put friend class CCoolBits; in XRootWindow)
=============================================================================*/

#include "CoolBitsPrivate.h" 

/*-----------------------------------------------------------------------------
	Implementations.
-----------------------------------------------------------------------------*/

IMPLEMENT_PACKAGE(CoolBits);

IMPLEMENT_CLASS(CCoolBits); 

IMPLEMENT_FUNCTION(CCoolBits,4004,execPreRenderWindows);
IMPLEMENT_FUNCTION(CCoolBits,4005,execPostRenderWindows);
IMPLEMENT_FUNCTION(CCoolBits,4006,execTraceTextureCoolBits);
IMPLEMENT_FUNCTION(CCoolBits,4007,execStriHash);
IMPLEMENT_FUNCTION(CCoolBits,4008,execBindConversationList);
IMPLEMENT_FUNCTION(CCoolBits,4009,execExportTravel);

/*-----------------------------------------------------------------------------
	PreRenderWindows() / PostRenderWindows()

	Meant as a replacement for the versions in PlayerPawnExt.
	A parameter is added to specify a drawning scale, e.g. you can
	force the windows to be renderer in 1x scale. However, if your
	screen is smaller then ( Scale * Width, Scale * Height ) the
	next lower valid Scale is used. Using Scale = 0 enables the
	default DeusEx behaviour, and enables you to have an Auto option.

	To use these you need to override your PlayerPawns PreRender() and
	PostRenderFlash() to make the use of these functions in place of the
	PlayerPawnExt ones.
-----------------------------------------------------------------------------*/

void CCoolBits::execPreRenderWindows(FFrame& Stack, RESULT_DECL)
{
	guard( CCoolBits::execPreRenderWindows );

	P_GET_OBJECT( XRootWindow, RootWindow )
	P_GET_OBJECT( UCanvas, Canvas );
	P_GET_INT( Scale );
	P_FINISH;

	PreRender( RootWindow, Canvas, Scale );

	unguard;
}

void CCoolBits::execPostRenderWindows(FFrame& Stack, RESULT_DECL)
{
	guard( CCoolBits::execPostRenderWindows );

	P_GET_OBJECT( XRootWindow, RootWindow )
	P_GET_OBJECT( UCanvas, Canvas);
	P_GET_INT( Scale );
	P_FINISH;

	PostRender( RootWindow, Canvas, Scale );

	unguard;
}

#define SCENENODE_STACK_MAX_SIZE 20

FSceneNode	SceneNodeStack[SCENENODE_STACK_MAX_SIZE];
INT					SceneNodeStackIndex = 0;

void CCoolBits::PreRender( XRootWindow* RootWindow, UCanvas* Canvas, INT Scale )
{
	if ( RootWindow )
	{
		ResizeRoot( RootWindow, Canvas, Scale );

		INT StartX, StartY;
		INT EndX, EndY;

		if ( Canvas && Canvas->Frame )
		{
			// Save Canvas->Frame
			if ( SceneNodeStackIndex < SCENENODE_STACK_MAX_SIZE )
				appMemcpy( &SceneNodeStack[SceneNodeStackIndex++], Canvas->Frame, sizeof(FSceneNode) );

			if ( RootWindow->bRender )
			{
				if ( RootWindow->bClipRender )
				{
					StartX = Canvas->Frame->XB + RootWindow->renderX * RootWindow->hMultiplier;
					StartY = Canvas->Frame->YB + RootWindow->renderY * RootWindow->vMultiplier;

					EndX = StartX + RootWindow->renderWidth	 * RootWindow->hMultiplier;
					EndY = StartY + RootWindow->renderHeight * RootWindow->vMultiplier;

					if ( StartX < Canvas->Frame->XB )
						StartX = Canvas->Frame->XB;

					if ( StartY < Canvas->Frame->YB )
						StartY = Canvas->Frame->YB;

					if ( EndX > ( Canvas->Frame->X + Canvas->Frame->XB ) )
						EndX = Canvas->Frame->X + Canvas->Frame->XB;

					if ( EndY > ( Canvas->Frame->Y + Canvas->Frame->YB ) )
						EndY = Canvas->Frame->Y + Canvas->Frame->YB;
		
					Canvas->Frame->XB = StartX;
					Canvas->Frame->YB = StartY;
					Canvas->Frame->X	= Canvas->Frame->X + Canvas->Frame->XB - StartX;
					Canvas->Frame->Y	= Canvas->Frame->Y + Canvas->Frame->YB - StartY;
				}
			}
			else
			{
				Canvas->Frame->X = 0;
				Canvas->Frame->Y = 0;
			}

			Canvas->Frame->ComputeRenderSize();

			Canvas->X			= Canvas->Frame->X;
			Canvas->Y			= Canvas->Frame->Y;
			Canvas->ClipX = Canvas->Frame->X;
			Canvas->ClipY = Canvas->Frame->Y;

			Canvas->SetClip( Canvas->Frame->XB, Canvas->Frame->YB, Canvas->Frame->X, Canvas->Frame->Y );
		}
	}
}

void CCoolBits::PostRender( XRootWindow* RootWindow, UCanvas* Canvas, INT Scale )
{
	FSceneNode TempNode;

	if ( RootWindow )
	{
		RootWindow->InitializeWindows();

		if ( Canvas && Canvas->Frame )
		{
			// Fill TempNode
			appMemcpy( &TempNode, Canvas->Frame, sizeof(FSceneNode) );

			RootWindow->rootFrame = &TempNode;

			// Restore Canvas->Frame
			if ( --SceneNodeStackIndex >= 0 )
				appMemcpy( Canvas->Frame, &SceneNodeStack[SceneNodeStackIndex], sizeof(FSceneNode) );

			Canvas->SetClip( 0, 0, Canvas->Frame->X, Canvas->Frame->Y );
		}

		RootWindow->PaintWindows( Canvas );

		RootWindow->rootFrame = NULL;
	}
}

void CCoolBits::ResizeRoot( XRootWindow* RootWindow, UCanvas* Canvas, INT Scale )
{
	INT hMult, vMult;

	if ( RootWindow )
	{
		if ( Canvas )
		{
			// calculate multipliers
			hMult = Canvas->ClipX / 640.0;
			vMult = Canvas->ClipY / 480.0;

			if ( hMult < 1 )
				hMult = 1;

			if ( vMult < 1 )
				vMult = 1;

			if ( hMult < vMult )
				vMult = hMult;
			else if ( hMult > vMult )
				hMult = vMult;

			// override scale
			if ( Scale > 0 && Scale < hMult && Scale < vMult )
			{
				hMult = Scale;
				vMult = Scale;
			}

			RootWindow->hMultiplier = hMult;
			RootWindow->vMultiplier = vMult;

			// reset some stuff
			RootWindow->x					= 0;
			RootWindow->y					= 0;
			RootWindow->hMargin0	= 0.0;
			RootWindow->hMargin1	= 0.0;
			RootWindow->vMargin0	= 0.0;
			RootWindow->vMargin1	= 0.0;
			RootWindow->winHAlign	= HALIGN_Left;
			RootWindow->winVAlign	= VALIGN_Top;

			RootWindow->hardcodedWidth = Canvas->ClipX / hMult;
			RootWindow->hardcodedHeight = Canvas->ClipY / vMult;

			if ( RootWindow->hardcodedWidth != RootWindow->width
				|| RootWindow->hardcodedHeight != RootWindow->height )
			{
				RootWindow->ConfigureChild( 0.0, 0.0, RootWindow->hardcodedWidth, RootWindow->hardcodedHeight );
			}

			RootWindow->ClampMousePosition();
		}
	}
}

/*-----------------------------------------------------------------------------
	TraceTextureCoolBits() - Fixed version of Actors.TraceTexture().

	However you need to add an Actor as the first parameter who's
	Level will be used and who's Location will be used as a default
	parameter for for Start.

	Example:
		Old: foreach TraceTexture(class'Actor', target, texName, texGroup, texFlags, HitLocation, HitNormal, EndTrace)
		New: foreach class'CoolBits'.static.TraceTextureCoolBits(Self, class'Actor', target, texName, texGroup, texFlags, HitLocation, HitNormal, EndTrace)
-----------------------------------------------------------------------------*/

// Find the FBspNode for Point.
INT CCoolBits::FindNode( UModel* Model, INT iPlane, FVector Point )
{
	guard( CCoolBits::FindNode );
	check(Model);

	INT Index = iPlane;

	while ( Index != INDEX_NONE )
	{
		FBspNode& Node = Model->Nodes(Index);

		if ( Node.iSurf != INDEX_NONE )
		{
			FBspSurf& Surf = Model->Surfs(Node.iSurf);

			//debugf( TEXT("FindNode: Testing: FBspNode(%i) with FBspSurf(%i).Texture %s."), Index, Node.iSurf, Surf.Texture ? Surf.Texture->GetName() : TEXT("None") );

			UBOOL Valid = 1;
			FLOAT OldPlaneDot = 0.0f;

			// Check for NumVertices == 0
			if ( Node.NumVertices )
			{
				for ( INT i = 0; i < Node.NumVertices && Valid; i++ )
				{
					// Get Points.
					FVector& A = Model->Points(Model->Verts(Node.iVertPool + i).pVertex);
					FVector& B = Model->Points(Model->Verts(Node.iVertPool + ((i+1) % Node.NumVertices)).pVertex);

					// Now build a plane.
					FPlane Plane( A, (B-A)^FVector( Node.Plane.X, Node.Plane.Y, Node.Plane.Z ) );

					// Now check Point against this plane.
					FLOAT PlaneDot = Plane.PlaneDot( Point );
					
						// Sign changed, check next FBspNode.
					if ( OldPlaneDot * PlaneDot < 0.0f )
					{
						Valid = 0;
					}
					// Sign stayed the same, everything fine.
					else
					{
						OldPlaneDot = PlaneDot;
					}
				}

				// Passed check.
				if ( Valid )
				{
					//debugf( TEXT("FindNode: Passed.") );
					return Index;
				}
				//else
				//{
					//debugf( TEXT("FindNode: Failed.") );					
				//}
			}
			//else
			//{
				//debugf( TEXT("FindNode: Skipping: Node.NumVertices == 0.") );
			//}
		}
		//else
		//{
			//debugf( TEXT("FindNode: Skipping: Node.iSurf == INDEX_NONE.") );
		//}

		Index = Node.iPlane;
	}
	
	return INDEX_NONE;
	unguard;
}

void CCoolBits::execTraceTextureCoolBits(FFrame& Stack, RESULT_DECL)
{
	guard( CCoolBits::execTraceTextureCoolBits );
	
	P_GET_ACTOR( Other );
	P_GET_OBJECT( UClass, BaseClass );
	P_GET_OBJECT_REF( AActor, Actor );
	P_GET_NAME_REF( TexName );
	P_GET_NAME_REF( TexGroup );
	P_GET_INT_REF( Flags );
	P_GET_VECTOR_REF( HitLoc );
	P_GET_VECTOR_REF( HitNorm );
	P_GET_VECTOR( End );
	P_GET_VECTOR_OPTX( Start, Other ? Other->Location : FVector( 0.0f, 0.0f, 0.0f ) );
	P_GET_VECTOR_OPTX( Extent, FVector( 0.0f, 0.0f, 0.0f ) );
	P_FINISH;

	if ( !Other )
	{
		debugf( TEXT("TraceTextureCoolBits() failed: Other is None.") );
		return;
	}

	FMemMark MemMark( GMem );

	if ( !BaseClass )
		BaseClass = AActor::StaticClass();

	FCheckResult* Res = Other->XLevel->MultiLineCheck( GMem, End, Start, Extent, 1, Other->Level, 0 );

	PRE_ITERATOR

		if ( Res )
		{
			*Actor		= Res->Actor;
			*HitLoc		= Res->Location;
			*HitNorm	= Res->Normal;

			*TexName	= NAME_None;
			*TexGroup	= NAME_None;
			*Flags		= 0;

			if ( Res->Actor == Other->Level )
			{
				Res->Item = FindNode( Other->XLevel->Model, Res->Item, Res->Location );

				if ( Res->Item != INDEX_NONE )
				{
					FBspNode& Node = Other->XLevel->Model->Nodes(Res->Item);

					if ( Node.iSurf != INDEX_NONE )
					{
						FBspSurf& Surf = Other->XLevel->Model->Surfs(Node.iSurf);

						if ( Surf.Texture )
						{
							*TexName	= Surf.Texture->GetFName();

							if ( Surf.Texture->GetOuter() )
							{
								*TexGroup	= Surf.Texture->GetOuter()->GetFName();
							}
						}

						*Flags = Surf.PolyFlags;
					}
				}
			}

			Res = (FCheckResult *)Res->Next;
		}
		else
		{
			Stack.Code = &Stack.Node->Script( wEndOffset + 1 );

			*Actor = NULL;
			break;
		}

	POST_ITERATOR

	MemMark.Pop();

	unguardexec;
}

/*-----------------------------------------------------------------------------
	StriHash() - Wrapper for appStrihash.
	
	Returns the case insensitive hash for a given string.
-----------------------------------------------------------------------------*/

void CCoolBits::execStriHash(FFrame& Stack, RESULT_DECL)
{
	guard(CCoolBits::execStriHash);
	
	P_GET_STR( Str );
	P_FINISH;

	*(DWORD *)Result = appStrihash( *Str );

	unguardexec;
}

/*-----------------------------------------------------------------------------
	BindConversationList()
	
	Can by used to bind conversation events to an arbitray actor.
	e.g. bind the DataLinks to your GameInfo. You need to set a
	proper BindName for your actor (e.g. JCDenton for DataLinks).
	
	Example:
	if ( MissionNumber >= 0 && MissionNumber <= 9 )
	{
		ConListClass = "DeusExConText.ConList_Mission0" $ MissionNumber;
	}
	else if ( MissionNumber >= 10 && MissionNumber <= 99 )
	{
		ConListClass = "DeusExConText.ConList_Mission" $ MissionNumber;
	}
	else
	{
		Log( "WTF IS THIS MISSION NUMBER " $ MissionNumber );
		return false;
	}
	Class'Coolbits'.Static.BindConversationList( ConListClass, Self );
-----------------------------------------------------------------------------*/

void CCoolBits::execBindConversationList(FFrame& Stack, RESULT_DECL)
{
	guard( CCoolBits::execBindConversationList );
	
	P_GET_STR( ConversationListClass );
	P_GET_ACTOR( A );
	P_FINISH;

	DConversationList*	ConvList = (DConversationList *)StaticLoadObject( DConversationList::StaticClass(), 0, *ConversationListClass, 0, LOAD_NoWarn, 0 );

	if ( ConvList )
	{
		ConvList->BindConversations( A );

		*(UBOOL *)Result = TRUE;
	}
	else
	{
		*(UBOOL *)Result = FALSE;	
	}

	unguardexec;
}

/*-----------------------------------------------------------------------------
	ExportTravel()
	
	Updates the Levels TravelInfo's entry about the given PlayerPawn.
	The name for the entry is PlayerPawn->PlayerReplicationInfo->PlayerName.
	You will probably want to call this in PlayerPawn.Destroy() before
	the inventory gets deleted.
	
	ULevel::SpawnPlayActor() will create and link the objects once the
	player joins again.
	You modify GameInfo.AcceptInventory(). An approach would be to check
	if the player has a NanoKeyRing and if not give the initial inventory
	and keep the old if the already has a NanoKeyRing.
	
	Warning:
		ION modified the ExportTravel() in a way that non Actor Objects
		will be include in the TravelInfo entry. And this is a serious
		problem since this is prone for crashes and strange issues.
		So be warned when using this. When it crashes your game, try
		getting rid of non actor export variables / objects. However
		the FlagBase is constructed in a way which the GC *really*
		like. However removing all references to flags and flagbases helps.
		TObjectIterator is helpful.
-----------------------------------------------------------------------------*/

void CCoolBits::execExportTravel(FFrame& Stack, RESULT_DECL)
{
	guard( AHXMutator::execExportTravel );
	
	P_GET_OBJECT( APlayerPawn, PlayerPawn );
	P_FINISH;

	if ( PlayerPawn 
		&& PlayerPawn->XLevel
		&& PlayerPawn->PlayerReplicationInfo
		&& PlayerPawn->PlayerReplicationInfo->PlayerName != TEXT("")
		)
	{
		GLog->Logf( TEXT("CoolBits Export travel for: %s"), *PlayerPawn->PlayerReplicationInfo->PlayerName );

		// ExportTravel player.
		FStringOutputDevice Out( TEXT("") );
		ExportTravel( Out, PlayerPawn );

		PlayerPawn->XLevel->TravelInfo.Set( PlayerPawn->PlayerReplicationInfo->PlayerName, *Out );
	}

	unguardexec;
}

void CCoolBits::ExportTravel( FStringOutputDevice& Out, UObject* RootObject )
{
	guard( ExportTravel );
	check( RootObject );

	TCHAR Temp[4096];
	
	TArray<UObject *> ObjectSet; // Type guessed
	
	ObjectSet.AddItem( RootObject );

	for ( INT i = 0; i < ObjectSet.Num(); i++ )
	{
		UObject* Object = ObjectSet(i);

		// HX_NOTE: Added check for Object
		if ( Object && !Object->IsA( AActor::StaticClass() ) || ((AActor *)Object)->bTravel )
		{
			Out.Logf( TEXT("Class=%s Name=%s Ptr=%08x\r\n{\r\n"), Object->GetClass()->GetPathName(), Object->GetName(), Object );

			for ( TFieldIterator<UProperty> It( Object->GetClass() ); It; ++It )
			{
				for ( INT i = 0; i < It->ArrayDim; i++ )
				{
					if ( ( It->PropertyFlags & CPF_Travel ) && It->ExportText( i, Temp, (BYTE *)Object, &Object->GetClass()->Defaults(0), 0 ) )
					{
						Out.Log( It->GetName() );

						if ( It->ArrayDim != 1 )
							Out.Logf( TEXT("[%i]"), i );

						Out.Log( TEXT("=") );

						UObjectProperty* ObjProp = Cast<UObjectProperty>( *It );

						if ( ObjProp && ObjProp->PropertyClass->IsChildOf( UObject::StaticClass() ) )
						{
							UObject* RefObj = *((UObject **)((BYTE *)Object + It->Offset + i * It->ElementSize));

							Out.Logf( TEXT("%08x %s\r\n"), RefObj, RefObj ? RefObj->GetName() : TEXT("None") );

							// HX_NOTE: Added check
							if ( RefObj )
								ObjectSet.AddUniqueItem( RefObj );
						}
						else
						{
							// Dunno, maybe for text of DeusExNote, etc.
							for ( TCHAR* Ch = Temp; *Ch; Ch++ )
								if ( *Ch == '\n' )
									*Ch = '\\';

							Out.Logf( TEXT("%s\r\n"), Temp );
						}
					}
				}
			}

			Out.Logf( TEXT("}\r\n") );
		}
	}

	unguard;
}

/*-----------------------------------------------------------------------------
	Autogenerate Names.
-----------------------------------------------------------------------------*/
/*
#define NAMES_ONLY
#define AUTOGENERATE_NAME(name) COOLBITS_API FName COOLBITS_##name;
#define AUTOGENERATE_FUNCTION(cls,idx,name) IMPLEMENT_FUNCTION(cls,idx,name)
#include "CoolBitsPrivate.h"
#undef AUTOGENERATE_FUNCTION
#undef AUTOGENERATE_NAME
#undef NAMES_ONLY
void RegisterNames()
{
	static INT Registered=0;
	if(!Registered++)
	{
		#define NAMES_ONLY
		#define AUTOGENERATE_NAME(name) extern COOLBITS_API FName COOLBITS_##name; COOLBITS_##name=FName(TEXT(#name),FNAME_Intrinsic);
		#define AUTOGENERATE_FUNCTION(cls,idx,name)
		#include "CoolBitsPrivate.h"
		#undef DECLARE_NAME
		#undef NAMES_ONLY
	}
}
*/
/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
